{{
@file pjdata.spin
Initialize JVM data and provide access to registers, etc....
}}
con
{{
         Stackframe format
         +---------------+
         | Local0        | <- JVM_FP
         | ...           |
         | LocalN        |
         +---------------+
         | Method MSB    | <- JVM_FP + numLocals
         | Method LSB    |
         +---------------+
         | Old FP MSB    |
         | Old FP LSB    |
         +---------------+
         | JVM_PC MSB    | <- PC for return
         | JVM_PC LSB    |
         +---------------+
         | Stack0        |
         | ....          |
         +---------------+
                           <- JVM_SP
}}
{{
         Memory map layout
  $0000  +---------------+ @javaProg
         |    header     |
  $0014  +---------------+
         | jemcodes      |
         | ...           |
  $XXX0  +---------------+ stackbase                 ---
         | Local0        | <- JVM_FP                  ^
         | ...           |                            |
         | LocalN        |                            |
         +---------------+                            |
         | Method MSB    | <- JVM_FP + numLocals      |
         | Method LSB    |                            |
         +---------------+                            |
         | Old FP MSB    |                            |
         | Old FP LSB    |
         +---------------+                       stack frame
         | JVM_PC MSB    |                            
         | JVM_PC LSB    |                            | 
         +---------------+                            |
         | Stack0        | operand stack              |
         | ....          |                            V
         +---------------+                           ---
         |               | <- JVM_SP
         |     free      |
         |               |
         |               |
         |     free      |
         |               |
         +---------------+
         |               | <- JVM_HP
         |   allocated   |
         |               |
         +---------------+
         |  static vars  | size := byte[@javaProg + JVM_STATIC]   'size in bytes
         |  each var     | if size == 0 
         |    2 bytes    |   size := 256
         +---------------+                           
  JAVASIZE                 heapbase (on javelin $7FFF)
  Header layout:
  
  0000:   dw      $A4     ; Placeholder for run/debug.
  0001:   dw      $0      ; Placeholder for the length.
  0002:   dw      $01     ; Version: 0.1
  0003:   dw      16      ; Number of static fields = 8 (maximum of 128 static fields)
  0004:   dw      $00     ; `main' method: Offset hi = $00
  0005:   dw      $2C     ;                       lo = $2C
  0006:   dw      $00     ; java.lang.Object class: Offset hi = $00
  0007:   dw      $8D     ;                                lo = $8D
  0008:   dw      $01     ; java.lang.String class: Offset hi = $01
  0009:   dw      $59     ;                                lo = $59
  000A:   dw      $00     ; Throwable.throwVMException method: Offset hi = $00
  000B:   dw      $52     ;                                           lo = $52
  000C:   dw      $09     ; String table offset hi = $09
  000D:   dw      $7F     ;                     lo = $7F
  000E:   dw      $00     ; Callback method offset hi = $00
  000F:   dw      $00     ;                        lo = $00
  0010:   dw      $01     ; java.lang.OutOfMemoryError class: Offset hi = $01
  0011:   dw      $2C     ;                                          lo = $2C
  0012:   dw      $09     ; Stack base hi = $09
  0013:   dw      $B0     ;            lo = $B0
  
}}
  'Javelin IDE constants
  kDebugMagic        = $1D
  kRunMagic          = $A4
  'activity values
  kPassive = 0   'JVM was running but has stopped (exception,end)
  kRun     = 1   'JVM running
  kJog     = 2   '???
  kBreak   = 3   'JVM reached breakpoint
  kEnd     = 4   'java program ended
obj
  pcn: "pjcon"
  'fs : "fat32"
var
  long  cog
  long  progptr
dat     ' don't change order of these variables
  cmd     long 0 ' bytecode instruction pointer
  ctrl    long 0 ' control command pointer
  jemp    long 0 ' bytecode pointer begin of jem code space
  pc      long 0 ' program counter
  bp      long 0 ' stack base pointer stack begin address
  sp      long 0 ' stack pointer current stack position always < heap end pointer
  fp      long 0 ' frame pointer current stack frame
  mp      long 0 ' method pointer keeps address of class method descriptor
  ap      long 0 ' allocated heap pointer for passing references from spin to pasm
  hp      long 0 ' heap end pointer so code knows what references are dynamic objects
  ep      long 0 ' exception pointer
  cp      long 0 ' carry pointer
  staticp long 0 ' static base pointer above heap keeps non-dynamic memory references
  strp    long 0 ' string const pointer non-dynamic table of strings
  tab     long 0 [4] ' tables for jvm engines
  utab    long 0 ' utility table pointer - unused ?
  ucmd    long 0 ' utility command pointer - unused ?
  udat    long 0 ' utility data pointer - unused ?
  
CON
  BREAKPOINTS = 10     'number of breakpoints supported by debugger
          
DAT
  javaProg       word 0                   'address of javaProg storage area
  javaSize       word 0                   'size of javaProg
  
  memSize        word 0                   'requested by QueryConfig
  progSize       word 0                   'requested by QueryConfig
  heapBase       word 0                   'requested by QueryConfig
  staticBase     word 0                   'base of static variables area
  stringIndex    word 0                   'address of string index table
  eeBase         word 0                   'requested by stamp\core\EEPROM.java
                                          '  set to return true available free eeprom space
  heapStart      word 0                   'for us
  bpList         word 0[BREAKPOINTS+1]    'list of breakpoints (entries are jvmPC address values)
  numBreakpoints byte 0                   'number of breakpoints
  activity       byte 0                   'requested by QueryStatus
  carry          byte 0                   'set to true if math expression overflows
  kComEndLoad    byte 0                   'holds device version ($50,$70,$78)
  notifyStepDone byte 0                   'set to true if jvm must notify IDE after step is done 
  mustDoStep     byte 0                   'set to true when jvm must do step
  eeByte         byte 0                   'to support eeprom read/write
  
          
PUB init(jprog,jsize,port,print) | i
{{ initialize jvm registers }}
  jemp  := jprog
  javaProg := jprog
  javaSize := jsize
{
  case port
    1: kComEndLoad := $78  'download via propeller TX/RX pins (no loopback, no echo)
    2: kComEndLoad := $70  'download via spin stamp SOUT/SIN pins (loopback and echo)
}
  'set some constant fields
  memSize := jsize               'total space for java program
  'progSize := $2000              'constant value, does not reflect true program size
  heapBase := (jsize+jprog) & $fff0      'highest even heap address
  eeBase := (heapBase - jsize)   'Set eeBase to reflect true free eeprom space
{
  'clear variables
  repeat i from 0 to BREAKPOINTS-1
    bpList[i] := 0
  numBreakpoints := 0
}
{  
  if debugger
    activity := kPassive
  else
    activity := kRun
  carry := false
  notifyStepDone := false
'}
  carry := 0
  cp := 0
  ep := 0
  ' initialize heap first ... first read before heap setup must add ProgOffset
  staticp := readJavaByte(pcn#JVM_STATIC+getProgOffset)  'statics allocated from heapBase
  if staticp == 0
    staticp += 256
  staticp := heapBase - staticp                    'HP starts just before statics
  hp    := staticp-2
  heapStart := hp
  strp  := readJavaIntBE(pcn#JVM_STRTABLE)         'mark start of string index table
  'initialize jvm registers
  mp := readJavaIntBE(pcn#JVM_MAIN)                   'first method is 'main'
  pc := readJavaIntBE(mp+pcn#JVM_CODE)                'Get address for 'main' code
  bp := readJavaIntBE(pcn#JVM_STKBASE)                'SB loaded with stackbase
  'clear heap/stack area
  'repeat i from bp to memSize-1
  '  writeJavaByte(i,0)
  bytefill(bp,memSize-1,0)
  fp := bp                                        'FP starts at initial stackframe
  sp := fp                                        'SP starts at initial stackframe
  sp += (readJavaByte(mp+pcn#JVM_PARAMETERS)&$7F) 'skip parameters (b7 set if exception table present)
  sp += readJavaByte(mp + pcn#JVM_LOCALS)         'skip locals
  writeJavaIntBE(sp+0,$FFFF)                       'old mp
  writeJavaIntBE(sp+2,$0000)                       'old fp
  writeJavaIntBE(sp+4,$0006)                       'old pc (= Object method ???)
  sp += 6                                          'skip saved pointers
  
PUB writeJavaByte(address,value)
{{ write value to java program }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  byte[javaProg + address] := value
  
PUB readJavaByte(address):value
{{ read value from java program }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  value := byte[javaProg + address] & $FF
  'else
  '  value := pread() .... how to read from jem.bin file ????
PUB writeJavaIntBE(address,value)
{{ write value to java program using big endian }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  byte[javaProg + address] := value.byte[1]
  byte[javaProg + address + 1] := value.byte[0]
  
PUB readJavaIntBE(address):value
{{ read value from java program using big endian }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  value := byte[javaProg + address + 1]
  value.byte[1] := byte[javaProg + address]
  value &= $FFFF
            
pub getTablePtr(n)
  return @tab+4*n
pub getUtilPtr
  return @utab
pub getCmdPtr
  return @cmd
pub getControlPtr
  return @ctrl
    
pub getPCPtr
  return @pc
pub getPC
  return pc
pub getMP
  return mp
pub getFP
  return fp
pub getBP
  return bp
pub getSB
  return bp
pub getSP
  return sp
pub getHS
  return heapStart
pub getHP
  return hp
pub setHP(val)
  hp := val
pub getAP
  return ap
pub setAP(val)
  ap := val
pub getHB
  return heapBase
pub getHeapBase
  return heapBase
pub getEP
  return ep
pub getSTRP
  return strp
pub getStaticPtr
  return staticp
pub incrPC(val)
  pc += val
pub getProgOffset
  return javaProg
pub getRefHB
  return @heapBase
pub getRefHS
  return @heapStart
pub getRefStaticPtr
  return @staticp
pub getRefHP
  return @hp
pub getRefSP
  return @sp
  
  
'_____________________________________________________________________________
'
'push, pop, writeJVMregister, readJCMregister, etc... are for native support only.
'_____________________________________________________________________________
'
CON
  nmPortA       = $06  'corresponds to I/O pins 0..7
  nmPortB       = $07  'corresponds to I/O pins 8..15
  'reserved values for extending Javelin code to PropJava code
  nmPortC       = $08  'corresponds to I/O pins 16..23
  nmPortD       = $09  'corresponds to I/O pins 24..31
  
  pin0  = (1<<0)         'A constant representing pin P0
  pin1  = (1<<1)         'A constant representing pin P1
  pin2  = (1<<2)         'A constant representing pin P2
  pin3  = (1<<3)         'A constant representing pin P3 
  pin4  = (1<<4)         'A constant representing pin P4 
  pin5  = (1<<5)         'A constant representing pin P5
  pin6  = (1<<6)         'A constant representing pin P6 
  pin7  = (1<<7)         'A constant representing pin P7 
  pin8  = (1<<0) + $100  'A constant representing pin P8
  pin9  = (1<<1) + $100  'A constant representing pin P9
  pin10 = (1<<2) + $100  'A constant representing pin P10
  pin11 = (1<<3) + $100  'A constant representing pin P11 
  pin12 = (1<<4) + $100  'A constant representing pin P12 
  pin13 = (1<<5) + $100  'A constant representing pin P13
  pin14 = (1<<6) + $100  'A constant representing pin P14
  pin15 = (1<<7) + $100  'A constant representing pin P15  
  'reserved values for extending Javelin code to PropJava code
  pin16 = (1<<0) + $200  'A constant representing pin P16
  pin17 = (1<<1) + $200  'A constant representing pin P17
  pin18 = (1<<2) + $200  'A constant representing pin P18
  pin19 = (1<<3) + $200  'A constant representing pin P19 
  pin20 = (1<<4) + $200  'A constant representing pin P20 
  pin21 = (1<<5) + $200  'A constant representing pin P21
  pin22 = (1<<6) + $200  'A constant representing pin P22
  pin23 = (1<<7) + $200  'A constant representing pin P23  
  pin24 = (1<<0) + $300  'A constant representing pin P24
  pin25 = (1<<1) + $300  'A constant representing pin P25
  pin26 = (1<<2) + $300  'A constant representing pin P26
  pin27 = (1<<3) + $300  'A constant representing pin P27 
  pin28 = (1<<4) + $300  'A constant representing pin P28 
  pin29 = (1<<5) + $300  'A constant representing pin P29
  pin30 = (1<<6) + $300  'A constant representing pin P30
  pin31 = (1<<7) + $300  'A constant representing pin P31  
  'native registers
  'file stamp\core\EEPROM.java
  JVM_SB_MSB = $4C  '// Highbyte of stackbase (first free eeprom address)
  JVM_SB_LSB = $4D  '// Lowbyte of stackbase (first free eeprom address)
  {{
     The EEPROM.java methods size(),write() and read() use $7FFF for eeprom size.
     On the prop the eeprom size is set to JAVASIZE. To let the methods access the
     correct addresses, we must adjust for this difference. That means the value
     from register JVM_SB must be increased by JVM_SB_ADJUST. Originally JVM_SB is
     set to the stack base as found on eeprom address $0012.
  }}
  'file stamp\core\Memory.java
  JVM_HEAP_POINTER_LSB  = $18   '//jvm heap pointer (word)
  JVM_HEAP_POINTER_MSB  = $19
  JVM_STACK_POINTER_LSB = $12   '//jvm stack pointer (word)
  JVM_STACK_POINTER_MSB = $13
  
PUB push(value)
  writeJavaIntBE(sp,value)
  sp+=2
  
PUB pop: value
  sp-=2
  return readJavaIntBE(sp) 
    
PUB writeJVMregister(address,value)
{{ write value to register }}
  address &= $FF
  case address
    nmPortA: outa[7..0] := value
    nmPortB: outa[15..8] := value
    nmPortC: outa[23..16] := value
    nmPortD: outa[31..24] := value
  
PUB readJVMregister(address):value
{{ read value from register }}
  value := 0
  address &= $FF
  case address
    nmPortA: value := ina
    nmPortB: value := ina >> 8
    nmPortC: value := ina >> 16
    nmPortD: value := ina >> 24
    JVM_SB_MSB: value := (($8000 - javaSize) + bp) >> 8  '// Highbyte of simulated stackbase (first free eeprom address)
    JVM_SB_LSB: value := (($8000 - javaSize) + bp)       '// Lowbyte of simulated stackbase (first free eeprom address)
    JVM_HEAP_POINTER_MSB: value := hp >> 8  '//jvm heap pointer (highbyte)
    JVM_HEAP_POINTER_LSB: value := hp       '//jvm heap pointer (lowbyte)
    JVM_STACK_POINTER_MSB: value := sp >> 8 '//jvm stack pointer (highbyte)
    JVM_STACK_POINTER_LSB: value := sp      '//jvm stack pointer (lowbyte)
  value &= $FF
pub isCarry
  return cp
pub setCarry
  cp := 1
  
pub getProgBase
  return javaProg
pub getMemSize
  return javaSize

{
obj ee : "PasmEEPROM" ' need to test this !!!

PUB eeBlockWrite(address)
{{ write block of 16 bytes from javaProg to eeprom }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  ee.writeBuffer(ee#EEID, javaProg+address, 16, javaProg+address+16)
' need ee write code here

PUB eeWriteByte(address,value)
{{ write byte to javaProg eeprom }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  eeByte := value
  ee.writeByte(javaProg+address, value)

PUB eeReadByte(address)
{{ read byte from javaProg eeprom }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  eeByte := ee.readByte(address)
  eeByte &= $ff
  return eeByte
'}

'{
obj ee : "jvmEEPROM"
PUB eeBlockWrite(address)
{{ write block of 16 bytes from javaProg to eeprom }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  ee.VarBackup(javaProg+address,javaProg+address+15)
PUB eeWriteByte(address,value)
{{ write byte to javaProg eeprom }}
  if (address > getHP)
    address -= getProgOffset    ' subtract program offset after HP setup
  eeByte := value
  ee.FromRam(@eeByte,@eeByte,javaProg+address)
PUB eeReadByte(address): value
{{ read byte from javaProg eeprom }}
  ee.ToRam(@eeByte,@eeByte,javaProg+address)
  return eeByte & $FF
'}

{
'' extra stuff for jvmComm support
''
PUB debugger: YesNo
{{ return true if javaprogram will be debugged }}
  return byte[javaProg] == kDebugMagic
PUB setActivity(kActivity)
{{ set jvm state to kPassive,kRun,kJog,kBreak }}
  activity := kActivity
PUB getActivity: kActivity
{{ get current activity state }}
  return activity & $FF
PUB getProgSize: value
{{ get program size }}
  return progSize & $FFFF
PUB setMustDoStep
{{ set must do step }}
  mustDoStep := $FF
PUB clearMustdoStep
{{ clear must do step }}
  mustDoStep := $00
PUB mustStep: YesNo
{{ test if must do step }}
  return mustdoStep == $FF
PUB setNotifyStepDone
{{ set step }}
  notifyStepDone := $FF
PUB clearNotifyStepDone
{{ clear step }}
  notifyStepDone := $00
PUB mustNotifyStepDone: YesNo
{{ test notifyStepDone }}
  return notifyStepDone == $FF
PUB setBreakpointByte(index,value)
{{ write byte to bpList }}
  byte[@bpList + index] := value
  
PUB setBreakpoints(value)
{{ set number of breakpoints }}
  numBreakpoints := value
PUB isBreakpoint: YesNo | i
{{ check if pc is a breakpoint }}
  i := 0
  repeat numBreakpoints
    if bpList[i] => PC-3 AND bpList[i] < PC 'breakpoints are start addresses of a jemcode instruction
      return true                                 'and we want to stop AFTER the instruction
    i++
  return false
'}     
