{{
@file pjvm_demo.spin:

  This is a PASM version of Peter's Spin JVM.
  This code is compatible with much of the Javelin Stamp
  library at http://tech.groups.yahoo.com/group/JavelinCode
  (testing is still in progress, so there may be errors.)

  '------------------------------------------------
  The JVM provides 18KB to 19KB for user Java bytecode.
  This is different from the Javelin which has 32KB.
  '------------------------------------------------

  '------------------------------------------------
  In the future, being able to define custom JNI devices
  is very desirable. LCD, touch screen, micro-ball mouse,
  keyboard/keypad, tone generator, etc... are possible.
  '------------------------------------------------

  '------------------------------------------------
  Minimum COG configuration:
  2 COGs used; 6 COGs free for user apps.

  1 for Spin and VM service handler.
  1 for JVM Core Interpreter.
  '------------------------------------------------

  '------------------------------------------------
  Standard Javelin-like configuration:
  5 COGs used; 3 free for user apps.

  1 for Spin and VM service handler.
  1 for JVM Core Interpreter.
  3 for 6 VPs by default with up to 4 for 8 VPs possible.
  '------------------------------------------------

  '------------------------------------------------
  Other Notes:
  '------------------------------------------------
  Use FDX define to enable high speed serial - uses 1 COG
  Garbage collection is not functional now and disabled.
  '------------------------------------------------

  '------------------------------------------------
  Performance:
  '------------------------------------------------
  A recursive FIBO test gives the following with 80MHz clock

  FIBO(00) =     1 (    7 ticks) (    4 ms)
  FIBO(01) =     1 (    7 ticks) (    4 ms)
  FIBO(02) =     1 (    7 ticks) (    4 ms)
  FIBO(03) =     2 (    7 ticks) (    4 ms)
  FIBO(04) =     3 (    8 ticks) (    4 ms)
  FIBO(05) =     5 (    8 ticks) (    4 ms)
  FIBO(06) =     8 (    9 ticks) (    5 ms)
  FIBO(07) =    13 (    9 ticks) (    5 ms)
  FIBO(08) =    21 (   11 ticks) (    6 ms)
  FIBO(09) =    34 (   14 ticks) (    8 ms)
  FIBO(10) =    55 (   19 ticks) (   11 ms)
  FIBO(11) =    89 (   26 ticks) (   14 ms)
  FIBO(12) =   144 (   43 ticks) (   24 ms)
  FIBO(13) =   233 (   59 ticks) (   33 ms)
  FIBO(14) =   377 (   91 ticks) (   51 ms)
  FIBO(15) =   610 (  143 ticks) (   79 ms)
  FIBO(16) =   987 (  227 ticks) (  126 ms)
  FIBO(17) =  1597 (  364 ticks) (  202 ms)
  FIBO(18) =  2584 (  585 ticks) (  325 ms)
  FIBO(19) =  4181 (  943 ticks) (  524 ms)
  FIBO(20) =  6765 ( 1520 ticks) (  844 ms)
  FIBO(21) = 10946 ( 2456 ticks) ( 1364 ms)
  FIBO(22) = 17711 ( 3969 ticks) ( 2205 ms)
  FIBO(23) = 28657 ( 6419 ticks) ( 3566 ms)
  FIBO(24) = 46368 (10381 ticks) ( 5767 ms)
  '------------------------------------------------

}}

con
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  '_clkmode = xtal1 + pll8x
  '_xinfreq = 10_000_000

obj
#ifdef FDX
  sx : "FullDuplexSingleton"
#else
  sx : "SingleSerial"
#endif
  pj : "pjvmInterpreterCOG"
  pjd: "pjdata"
  pcn: "pjcon"
  pn : "pnative"
#ifdef GCOLLECT
  pm : "pjmemory"               ' best fit list power of 2 size
#endif

pub main | progptr, ctrlp, cmdptr
{{
Startup the JVM interpreter
}}
#ifdef FDX
  sx.start(31,30,0,19200)
#else
  sx.init(31,30,19200)
#endif
  waitcnt(clkfreq/1+cnt)
  'sx.str(string($d,"Hello JVM",$d))

  ctrlp := pjd.getControlPtr
  cmdptr := pjd.getCmdPtr

  progptr := pj.getProgPtr
  longmove(pj.getProgBuffer, progptr, pj.getProgSize/4)
  progptr := pj.getProgBuffer
  
  pjd.init(progptr,pj.getJavaCodeSize,1,false)
#ifdef GCOLLECT
  pm.init(pjd.getHS-2)
#endif

'  cognew(handler, @hstack)     ' for BMA debug only

  pn.start(3)                   ' start virtual peripherals
  pj.start(cmdptr)              ' start jvm engine

  handler

var long hstack[$40]

pub handler | cmd, tmp, cmdptr, ctlptr
{{
handle memory allocations, natives, and exceptions
}}
  cmdptr := pjd.getCmdPtr
  ctlptr := pjd.getControlPtr

  repeat

    if (long[ctlptr] > 0) and (long[ctlptr] & pj#MALLOC_BIT)

      tmp := long[ctlptr] & $3fff
#ifdef GCOLLECT
      pm.alloc(tmp)
      'pm.simplealloc(tmp)
#endif
      long[ctlptr] := -1

    if long[cmdptr] & pj#EXCPTN_BIT

      sx.str(string($d,"Exception "))
      sx.hex(long[cmdptr],4)
      sx.str(string(" at "))
      sx.hex(pjd.getPC,4)
      sx.tx($d)
      long[cmdptr] := 0
      repeat ' stall forever

    elseif long[cmdptr] & $100

      cmd := long[cmdptr] & $3f
      pn.doNative(cmd)
      pjd.incrPC(3)
      long[cmdptr] := 0


